@@ -1,7 +1,7 @@
 # -*- coding: utf-8 -*-
 
 # The MIT License (MIT)
-# Copyright (c) 2014-2021 Matias Bordese
+# Copyright (c) 2014-2022 Matias Bordese
 #
 # Permission is hereby granted, free of charge, to any person obtaining a copy
 # of this software and associated documentation files (the "Software"), to deal
@@ -38,7 +38,10 @@ from unidiff.constants import (
     LINE_TYPE_REMOVED,
     LINE_TYPE_NO_NEWLINE,
     LINE_VALUE_NO_NEWLINE,
+    RE_DIFF_GIT_DELETED_FILE,
     RE_DIFF_GIT_HEADER,
+    RE_DIFF_GIT_HEADER_NO_PREFIX,
+    RE_DIFF_GIT_NEW_FILE,
     RE_HUNK_BODY_LINE,
     RE_HUNK_EMPTY_BODY_LINE,
     RE_HUNK_HEADER,
@@ -52,8 +55,9 @@ from unidiff.errors import UnidiffParseError
 
 PY2 = sys.version_info[0] == 2
 if PY2:
+    import io
     from StringIO import StringIO
-    open_file = codecs.open
+    open_file = io.open
     make_str = lambda x: x.encode(DEFAULT_ENCODING)
 
     def implements_to_string(cls):
@@ -232,7 +236,7 @@ class PatchedFile(list):
 
     def __init__(self, patch_info=None, source='', target='',
                  source_timestamp=None, target_timestamp=None,
-                 is_binary_file=False, is_rename=False):
+                 is_binary_file=False):
         # type: (Optional[PatchInfo], str, str, Optional[str], Optional[str], bool, bool) -> None
         super(PatchedFile, self).__init__()
         self.patch_info = patch_info
@@ -241,7 +245,6 @@ class PatchedFile(list):
         self.target_file = target
         self.target_timestamp = target_timestamp
         self.is_binary_file = is_binary_file
-        self.is_rename = is_rename
 
     def __repr__(self):
         # type: () -> str
@@ -406,6 +409,12 @@ class PatchedFile(list):
         """Return the file total removed lines."""
         return sum([hunk.removed for hunk in self])
 
+    @property
+    def is_rename(self):
+        return (self.source_file != DEV_NULL
+            and self.target_file != DEV_NULL
+            and self.source_file[2:] != self.target_file[2:])
+
     @property
     def is_added_file(self):
         # type: () -> bool
@@ -470,20 +479,32 @@ class PatchSet(list):
                 line = line.decode(encoding)
 
             # check for a git file rename
-            is_diff_git_header = RE_DIFF_GIT_HEADER.match(line)
+            is_diff_git_header = RE_DIFF_GIT_HEADER.match(line) or RE_DIFF_GIT_HEADER_NO_PREFIX.match(line)
             if is_diff_git_header:
-                if patch_info is None:
-                    patch_info = PatchInfo()
+                patch_info = PatchInfo()
                 source_file = is_diff_git_header.group('source')
                 target_file = is_diff_git_header.group('target')
-                if (source_file != DEV_NULL
-                        and target_file != DEV_NULL
-                        and source_file[2:] != target_file[2:]):
-                    # this is a renamed file
-                    current_file = PatchedFile(
-                        patch_info, source_file, target_file, None, None,
-                        is_rename=True)
-                    self.append(current_file)
+                current_file = PatchedFile(
+                    patch_info, source_file, target_file, None, None)
+                self.append(current_file)
+                patch_info.append(line)
+                continue
+
+            # check for a git new file
+            is_diff_git_new_file = RE_DIFF_GIT_NEW_FILE.match(line)
+            if is_diff_git_new_file:
+                if current_file is None or patch_info is None:
+                    raise UnidiffParseError('Unexpected new file found: %s' % line)
+                current_file.source_file = DEV_NULL
+                patch_info.append(line)
+                continue
+
+            # check for a git deleted file
+            is_diff_git_deleted_file = RE_DIFF_GIT_DELETED_FILE.match(line)
+            if is_diff_git_deleted_file:
+                if current_file is None or patch_info is None:
+                    raise UnidiffParseError('Unexpected deleted file found: %s' % line)
+                current_file.target_file = DEV_NULL
                 patch_info.append(line)
                 continue
 
@@ -494,18 +515,20 @@ class PatchSet(list):
                 source_timestamp = is_source_filename.group('timestamp')
                 # reset current file, unless we are processing a rename
                 # (in that case, source files should match)
-                if current_file is not None and not (current_file.is_rename and
+                if current_file is not None and not (
                         current_file.source_file == source_file):
                     current_file = None
+                elif current_file is not None:
+                    current_file.source_timestamp = source_timestamp
                 continue
 
             # check for target file header
             is_target_filename = RE_TARGET_FILENAME.match(line)
             if is_target_filename:
-                if current_file is not None and not current_file.is_rename:
-                    raise UnidiffParseError('Target without source: %s' % line)
                 target_file = is_target_filename.group('filename')
                 target_timestamp = is_target_filename.group('timestamp')
+                if current_file is not None and not (current_file.target_file == target_file):
+                    raise UnidiffParseError('Target without source: %s' % line)
                 if current_file is None:
                     # add current file to PatchSet
                     current_file = PatchedFile(
@@ -513,6 +536,8 @@ class PatchSet(list):
                         source_timestamp, target_timestamp)
                     self.append(current_file)
                     patch_info = None
+                else:
+                    current_file.target_timestamp = target_timestamp
                 continue
 
             # check for hunk header
@@ -547,9 +572,12 @@ class PatchSet(list):
                 source_file = is_binary_diff.group('source_filename')
                 target_file = is_binary_diff.group('target_filename')
                 patch_info.append(line)
-                current_file = PatchedFile(
-                    patch_info, source_file, target_file, is_binary_file=True)
-                self.append(current_file)
+                if current_file is not None:
+                    current_file.is_binary_file = True
+                else:
+                    current_file = PatchedFile(
+                        patch_info, source_file, target_file, is_binary_file=True)
+                    self.append(current_file)
                 patch_info = None
                 current_file = None
                 continue
@@ -557,10 +585,10 @@ class PatchSet(list):
             patch_info.append(line)
 
     @classmethod
-    def from_filename(cls, filename, encoding=DEFAULT_ENCODING, errors=None):
+    def from_filename(cls, filename, encoding=DEFAULT_ENCODING, errors=None, newline=None):
         # type: (str, str, Optional[str]) -> PatchSet
         """Return a PatchSet instance given a diff filename."""
-        with open_file(filename, 'r', encoding=encoding, errors=errors) as f:
+        with open_file(filename, 'r', encoding=encoding, errors=errors, newline=newline) as f:
             instance = cls(f)
         return instance
 
